# Copyright 2025 Emcie Co Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import chain
from typing import Optional, Sequence, cast, Annotated, TypeAlias
from fastapi import APIRouter, HTTPException, Path, Query, status

from parlant.api import common
from parlant.api.common import (
    ExampleJson,
    GuidelineDTO,
    GuidelineIdField,
    RelationshipDTO,
    RelationshipKindDTO,
    TagDTO,
    TagIdField,
    ToolIdDTO,
    apigen_config,
    tool_to_dto,
)
from parlant.core.common import DefaultBaseModel
from parlant.core.relationships import (
    EntityType,
    GuidelineRelationshipKind,
    Relationship,
    RelationshipEntity,
    RelationshipId,
    RelationshipKind,
    RelationshipStore,
    ToolRelationshipKind,
)
from parlant.core.guidelines import Guideline, GuidelineId, GuidelineStore
from parlant.core.services.tools.service_registry import ServiceRegistry
from parlant.core.tags import Tag, TagId, TagStore
from parlant.api.common import relationship_example
from parlant.core.tools import Tool, ToolId

API_GROUP = "relationships"


relationship_creation_params_example: ExampleJson = {
    "source_guideline": "gid_123",
    "target_tag": "tid_456",
    "kind": "entailment",
}


relationship_creation_tool_example: ExampleJson = {
    "source_tool": {
        "service_name": "tool_service_name",
        "tool_name": "tool_name",
    },
    "target_tool": {
        "service_name": "tool_service_name",
        "tool_name": "tool_name",
    },
    "kind": "overlap",
}


class RelationshipCreationParamsDTO(
    DefaultBaseModel,
    json_schema_extra={
        "example": relationship_creation_params_example,
        "tool_example": relationship_creation_tool_example,
    },
):
    source_guideline: Optional[GuidelineIdField] = None
    source_tag: Optional[TagIdField] = None
    source_tool: Optional[ToolIdDTO] = None
    target_guideline: Optional[GuidelineIdField] = None
    target_tag: Optional[TagIdField] = None
    target_tool: Optional[ToolIdDTO] = None
    kind: RelationshipKindDTO


GuidelineIdQuery: TypeAlias = Annotated[
    GuidelineId,
    Query(description="The ID of the guideline to list relationships for"),
]


TagIdQuery: TypeAlias = Annotated[
    TagId,
    Query(description="The ID of the tag to list relationships for"),
]


IndirectQuery: TypeAlias = Annotated[
    bool,
    Query(description="Whether to include indirect relationships"),
]


RelationshipKindQuery: TypeAlias = Annotated[
    RelationshipKindDTO,
    Query(description="The kind of relationship to list"),
]


RelationshipIdPath: TypeAlias = Annotated[
    RelationshipId,
    Path(
        description="identifier of relationship",
        examples=[RelationshipId("gr_123")],
    ),
]


def _relationship_kind_to_dto(
    kind: RelationshipKind,
) -> RelationshipKindDTO:
    match kind:
        case GuidelineRelationshipKind.ENTAILMENT:
            return RelationshipKindDTO.ENTAILMENT
        case GuidelineRelationshipKind.PRIORITY:
            return RelationshipKindDTO.PRIORITY
        case GuidelineRelationshipKind.DEPENDENCY:
            return RelationshipKindDTO.DEPENDENCY
        case ToolRelationshipKind.OVERLAP:
            return RelationshipKindDTO.OVERLAP
        case _:
            raise ValueError(f"Invalid relationship kind: {kind.value}")


def _relationship_kind_dto_to_kind(
    dto: RelationshipKindDTO,
) -> RelationshipKind:
    match dto:
        case RelationshipKindDTO.ENTAILMENT:
            return GuidelineRelationshipKind.ENTAILMENT
        case RelationshipKindDTO.PRIORITY:
            return GuidelineRelationshipKind.PRIORITY
        case RelationshipKindDTO.DEPENDENCY:
            return GuidelineRelationshipKind.DEPENDENCY
        case RelationshipKindDTO.OVERLAP:
            return ToolRelationshipKind.OVERLAP
        case _:
            raise ValueError(f"Invalid relationship kind: {dto.value}")


def _get_relationship_entity(
    guideline_id: Optional[GuidelineId],
    tag_id: Optional[TagId],
    tool_id: Optional[ToolId],
) -> RelationshipEntity:
    if guideline_id:
        return RelationshipEntity(id=guideline_id, type=EntityType.GUIDELINE)
    elif tag_id:
        return RelationshipEntity(id=tag_id, type=EntityType.TAG)
    elif tool_id:
        return RelationshipEntity(id=tool_id, type=EntityType.TOOL)
    else:
        raise ValueError("No entity provided")


def create_router(
    guideline_store: GuidelineStore,
    tag_store: TagStore,
    relationship_store: RelationshipStore,
    service_registry: ServiceRegistry,
) -> APIRouter:
    async def relationship_to_dto(
        relationship: Relationship,
    ) -> RelationshipDTO:
        source_guideline = (
            await guideline_store.read_guideline(
                guideline_id=cast(GuidelineId, relationship.source.id)
            )
            if relationship.source.type == EntityType.GUIDELINE
            else None
        )

        source_tag = (
            await tag_store.read_tag(tag_id=cast(TagId, relationship.source.id))
            if relationship.source.type == EntityType.TAG
            else None
        )

        target_guideline = (
            await guideline_store.read_guideline(
                guideline_id=cast(GuidelineId, relationship.target.id)
            )
            if relationship.target.type == EntityType.GUIDELINE
            else None
        )

        target_tag = (
            await tag_store.read_tag(tag_id=cast(TagId, relationship.target.id))
            if relationship.target.type == EntityType.TAG
            else None
        )

        source_tool = (
            await (
                await service_registry.read_tool_service(
                    name=cast(ToolId, relationship.source.id).service_name
                )
            ).read_tool(name=cast(ToolId, relationship.source.id).tool_name)
            if relationship.source.type == EntityType.TOOL
            else None
        )

        target_tool = (
            await (
                await service_registry.read_tool_service(
                    name=cast(ToolId, relationship.target.id).service_name
                )
            ).read_tool(name=cast(ToolId, relationship.target.id).tool_name)
            if relationship.target.type == EntityType.TOOL
            else None
        )

        return RelationshipDTO(
            id=relationship.id,
            source_guideline=GuidelineDTO(
                id=cast(Guideline, source_guideline).id,
                condition=cast(Guideline, source_guideline).content.condition,
                action=cast(Guideline, source_guideline).content.action,
                enabled=cast(Guideline, source_guideline).enabled,
                tags=cast(Guideline, source_guideline).tags,
                metadata=cast(Guideline, source_guideline).metadata,
            )
            if relationship.source.type == EntityType.GUIDELINE
            else None,
            source_tag=TagDTO(
                id=cast(Tag, source_tag).id,
                name=cast(Tag, source_tag).name,
            )
            if relationship.source.type == EntityType.TAG
            else None,
            target_guideline=GuidelineDTO(
                id=cast(Guideline, target_guideline).id,
                condition=cast(Guideline, target_guideline).content.condition,
                action=cast(Guideline, target_guideline).content.action,
                enabled=cast(Guideline, target_guideline).enabled,
                tags=cast(Guideline, target_guideline).tags,
                metadata=cast(Guideline, target_guideline).metadata,
            )
            if relationship.target.type == EntityType.GUIDELINE
            else None,
            target_tag=TagDTO(
                id=cast(Tag, target_tag).id,
                name=cast(Tag, target_tag).name,
            )
            if relationship.target.type == EntityType.TAG
            else None,
            source_tool=tool_to_dto(cast(Tool, source_tool))
            if relationship.source.type == EntityType.TOOL
            else None,
            target_tool=tool_to_dto(cast(Tool, target_tool))
            if relationship.target.type == EntityType.TOOL
            else None,
            indirect=False,
            kind=_relationship_kind_to_dto(cast(GuidelineRelationshipKind, relationship.kind)),
        )

    router = APIRouter()

    @router.post(
        "",
        status_code=status.HTTP_201_CREATED,
        operation_id="create_relationship",
        response_model=RelationshipDTO,
        responses={
            status.HTTP_201_CREATED: {
                "description": "Relationship successfully created. Returns the created relationship.",
                "content": common.example_json_content(relationship_example),
            },
            status.HTTP_422_UNPROCESSABLE_ENTITY: {
                "description": "Validation error in request parameters"
            },
        },
        **apigen_config(group_name=API_GROUP, method_name="create"),
    )
    async def create_relationship(
        params: RelationshipCreationParamsDTO,
    ) -> RelationshipDTO:
        """
        Create a relationship.

        A relationship is a relationship between a guideline and a tag.
        It can be created between a guideline and a tag, or between two guidelines, or between two tags.
        """
        if params.source_guideline and params.source_tag:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="A relationship cannot have both a source guideline and a source tag",
            )
        elif params.target_guideline and params.target_tag:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="A relationship cannot have both a target guideline and a target tag",
            )
        elif (
            params.source_guideline
            and params.target_guideline
            and params.source_guideline == params.target_guideline
        ) or (params.source_tag and params.target_tag and params.source_tag == params.target_tag):
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="source and target cannot be the same entity",
            )

        if params.source_guideline:
            await guideline_store.read_guideline(guideline_id=params.source_guideline)
        elif params.source_tag:
            await tag_store.read_tag(tag_id=params.source_tag)
        elif params.source_tool:
            service = await service_registry.read_tool_service(name=params.source_tool.service_name)
            _ = await service.read_tool(name=params.source_tool.tool_name)

        if params.target_guideline:
            await guideline_store.read_guideline(guideline_id=params.target_guideline)
        elif params.target_tag:
            await tag_store.read_tag(tag_id=params.target_tag)
        elif params.target_tool:
            service = await service_registry.read_tool_service(name=params.target_tool.service_name)
            _ = await service.read_tool(name=params.target_tool.tool_name)

        relationship = await relationship_store.create_relationship(
            source=_get_relationship_entity(
                params.source_guideline,
                params.source_tag,
                ToolId(params.source_tool.service_name, params.source_tool.tool_name)
                if params.source_tool
                else None,
            ),
            target=_get_relationship_entity(
                params.target_guideline,
                params.target_tag,
                ToolId(params.target_tool.service_name, params.target_tool.tool_name)
                if params.target_tool
                else None,
            ),
            kind=_relationship_kind_dto_to_kind(params.kind),
        )

        return await relationship_to_dto(relationship=relationship)

    @router.get(
        "",
        operation_id="list_relationships",
        response_model=Sequence[RelationshipDTO],
        responses={
            status.HTTP_200_OK: {
                "description": "Relationships successfully retrieved. Returns a list of all relationships.",
                "content": common.example_json_content([relationship_example]),
            },
        },
        **apigen_config(group_name=API_GROUP, method_name="list"),
    )
    async def list_relationships(
        kind: RelationshipKindQuery,
        indirect: IndirectQuery = True,
        guideline_id: Optional[GuidelineIdQuery] = None,
        tag_id: Optional[TagIdQuery] = None,
    ) -> Sequence[RelationshipDTO]:
        """
        List relationships.

        Either `guideline_id` or `tag_id` must be provided.
        """
        if guideline_id is None and tag_id is None:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Either guideline_id or tag_id must be provided",
            )

        if guideline_id:
            await guideline_store.read_guideline(guideline_id=guideline_id)
        elif tag_id:
            await tag_store.read_tag(tag_id=tag_id)

        entity_id = (
            cast(GuidelineId | TagId, guideline_id)
            if guideline_id
            else cast(GuidelineId | TagId, tag_id)
        )

        source_relationships = await relationship_store.list_relationships(
            kind=_relationship_kind_dto_to_kind(kind),
            source_id=entity_id,
            indirect=indirect,
        )

        target_relationships = await relationship_store.list_relationships(
            kind=_relationship_kind_dto_to_kind(kind),
            target_id=entity_id,
            indirect=indirect,
        )
        relationships = chain(source_relationships, target_relationships)

        return [
            await relationship_to_dto(relationship=relationship) for relationship in relationships
        ]

    @router.get(
        "/{relationship_id}",
        operation_id="read_relationship",
        status_code=status.HTTP_200_OK,
        response_model=RelationshipDTO,
        responses={
            status.HTTP_200_OK: {
                "description": "Relationship successfully retrieved. Returns the requested relationship.",
                "content": common.example_json_content(relationship_example),
            },
        },
        **apigen_config(group_name=API_GROUP, method_name="retrieve"),
    )
    async def read_relationship(
        relationship_id: RelationshipIdPath,
    ) -> RelationshipDTO:
        """
        Read a relationship by ID.
        """
        relationship = await relationship_store.read_relationship(id=relationship_id)

        return await relationship_to_dto(relationship=relationship)

    @router.delete(
        "/{relationship_id}",
        operation_id="delete_relationship",
        status_code=status.HTTP_204_NO_CONTENT,
        responses={
            status.HTTP_204_NO_CONTENT: {"description": "Relationship successfully deleted."},
            status.HTTP_404_NOT_FOUND: {"description": "Relationship not found."},
        },
        **apigen_config(group_name=API_GROUP, method_name="delete"),
    )
    async def delete_relationship(
        relationship_id: RelationshipIdPath,
    ) -> None:
        """
        Delete a relationship by ID.
        """
        await relationship_store.delete_relationship(id=relationship_id)

    return router
